#ifndef FSGUICOLORDLG_IS_INCLUDED
#define FSGUICOLORDLG_IS_INCLUDED
/* { */

#include "../fsguidialog.h"

class FsGuiColorDialog : public FsGuiDialog
{
public:
	FsGuiColorPalette *linkedColorPalette;
	YsColor currentColor;

	class CurrentColor : public FsGuiDialogItem
	{
	private:
		CurrentColor &operator=(const CurrentColor &){}
	public:
		YsColor &showColor;

		CurrentColor(YsColor &currentColor) : showColor(currentColor)
		{
		}
		virtual void Draw(YSBOOL focus,YSBOOL mouseOver,const YsColor &defBgCol,const YsColor &defFgCol,const YsColor &activeBgCol,const YsColor &activeFgCol,const YsColor &frameCol) const;
	};
	class ColorBand : public FsGuiDialogItem
	{
	private:
		ColorBand &operator=(const ColorBand &){}
	public:
		double s,v;
		YsColor &showColor;

		ColorBand(YsColor &currentColor) : showColor(currentColor)
		{
		}
		virtual void Draw(YSBOOL focus,YSBOOL mouseOver,const YsColor &defBgCol,const YsColor &defFgCol,const YsColor &activeBgCol,const YsColor &activeFgCol,const YsColor &frameCol) const;
		virtual YSRESULT KeyIn(int key,YSBOOL shift,YSBOOL ctrl,YSBOOL alt);
		virtual void SetMouseState(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	};
	class Brightness : public FsGuiDialogItem
	{
	private:
		Brightness &operator=(const Brightness &){}
	public:
		YsColor &showColor;

		Brightness(YsColor &currentColor) : showColor(currentColor)
		{
		}

		virtual void Draw(YSBOOL focus,YSBOOL mouseOver,const YsColor &defBgCol,const YsColor &defFgCol,const YsColor &activeBgCol,const YsColor &activeFgCol,const YsColor &frameCol) const;
		virtual void SetMouseState(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	};

private:
	FsGuiButton *closeBtn;
	CurrentColor *curColorStatic; // <- showColor is the selected color.
	ColorBand *colorBand;
	Brightness *brightness;

	FsGuiTextBox *redText,*greenText,*blueText;
	FsGuiTextBox *hueText,*saturationText,*valueText;

	FsGuiSlider *hueSlider;
	FsGuiSlider *saturationSlider;
	FsGuiSlider *valueSlider;

	FsGuiSlider *redSlider;
	FsGuiSlider *greenSlider;
	FsGuiSlider *blueSlider;

	FsGuiButton *setToPaletteBtn[16];

	FsGuiColorDialog();
	~FsGuiColorDialog();

public:
	static FsGuiColorDialog *Create(void);
	static void Delete(FsGuiColorDialog *ptr);

	void Make(FsGuiColorPalette *linkedColorPalette);
	void SetCurrentColor(const YsColor &col);
private:
	void UpdateRGBSliderFromCurrentColor(void);
	void UpdateHSVSliderFromCurrentColor(void);
	void UpdateTextBoxFromSlider(void) const;
	YsColor CaptureRGBSlider(void) const;
	YsColor CaptureHSVSlider(void) const;

	void UpdateRGBTextFromCurrentColor(void);
	void UpdateHSVTextFromCurrentColor(void);
	void UpdateSliderFromText(void);
	YsColor CaptureRGBText(void) const;
	YsColor CaptureHSVText(void) const;

public:
	const YsColor GetCurrentColor(void) const;

	virtual void OnButtonClick(FsGuiButton *btn);
	virtual void OnSliderPositionChange(FsGuiSlider *slider,const double &prevPos,const double &prevValue);
	virtual void OnTextBoxChange(FsGuiTextBox *txt);
};

/* } */
#endif
