/* ////////////////////////////////////////////////////////////

  File Name: edit_roundmenu.cpp
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#include "../fsgui3dapp.h"
#include "../miscdlg.h"
#include <ysgl.h>

#include <ysshellextedit_roundutil.h>
#include <ysshellextedit_offsetutil.h>
#include <ysshellextedit_topologyutil.h>
#include <ysshellextedit_localop.h>
#include <ysshellext_trackingutil.h>
#include "edit_rounddialog.h"

#include "../textresource.h"

PolyCreRoundDialog::PolyCreRoundDialog()
{
	okBtn=NULL;
	cancelBtn=NULL;
	slider=NULL;
	subdivTxt=NULL;
}

PolyCreRoundDialog::~PolyCreRoundDialog()
{
}

/*static*/ PolyCreRoundDialog *PolyCreRoundDialog::Create(class FsGui3DMainCanvas *canvas)
{
	PolyCreRoundDialog *newDlg=new PolyCreRoundDialog;
	newDlg->canvas=canvas;
	return newDlg;
}

/*static*/ void PolyCreRoundDialog::Delete(PolyCreRoundDialog *ptr)
{
	delete ptr;
}


YSRESULT PolyCreRoundDialog::Make(const YsShell *shl)
{
	this->shl=shl;

	if(NULL==okBtn)
	{
		Initialize();

		okBtn=AddTextButton(0,FSKEY_SPACE,FSGUI_PUSHBUTTON,FSGUI_COMMON_OK,YSTRUE);
		cancelBtn=AddTextButton(0,FSKEY_ESC,FSGUI_PUSHBUTTON,FSGUI_COMMON_CANCEL,YSFALSE);

		slider=AddHorizontalSlider(0,FSKEY_NULL,50,0.0,1.0,YSTRUE);

		subdivTxt=AddTextBox(0,FSKEY_NULL,FSGUI_DLG_COMMON_DIVCOUNT,L"2",5,YSTRUE);
	}

	slider->SetPosition(0.4);

	SetArrangeType(FSDIALOG_ARRANGE_TOP_LEFT);
	SetBackgroundAlpha(0.3);
	Fit();

	canvas->ArrangeDialog();

	return YSOK;
}


void PolyCreRoundDialog::OnButtonClick(FsGuiButton *btn)
{
	if(btn==okBtn)
	{
		if(YSOK==canvas->Edit_Round_SpaceKeyCallBack())
		{
			canvas->Edit_ClearUIMode();
		}
	}
	else if(btn==cancelBtn)
	{
		canvas->Edit_ClearUIMode();
	}
}

void PolyCreRoundDialog::OnSliderPositionChange(FsGuiSlider *slider,const double &,const double &)
{
	const double radius=maxRadius*slider->GetPosition();
	const int nDiv=YsBound(subdivTxt->GetInteger(),1,256);

	if(YSTRUE==mode3D)
	{
		roundUtil3d.CalculateRoundingAll(*shl,radius);
		roundUtil3d.SubdivideByNumberOfSubdivision(nDiv);
	}
	else
	{
		roundUtil2d.CalculateRoundingAll(*shl,radius,nDiv);
	}
}


////////////////////////////////////////////////////////////

/*static*/ void FsGui3DMainCanvas::Edit_Round_PolygonOrConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	canvas->Edit_Round_PolygonOrConstEdge(); // Test Implementation
}

void FsGui3DMainCanvas::Edit_Round_PolygonOrConstEdge(void)
{
	if(NULL!=slHd)
	{
		YsShellExtEdit::StopIncUndo incUndo(slHd);

		YsArray <YsShellVertexHandle> selVtHd;
		slHd->GetSelectedVertex(selVtHd);

		YsArray <YsShellPolygonHandle> selPlHd;
		slHd->GetSelectedPolygon(selPlHd);

		YsArray <YsShellExt::ConstEdgeHandle> selCeHd;
		slHd->GetSelectedConstEdge(selCeHd);

		roundDlg->roundUtil2d.CleanUp();

		for(auto plHd : selPlHd)
		{
			roundDlg->roundUtil2d.SetUpRoundPolygon(*(const YsShell *)slHd,plHd,NULL);
		}

		for(auto ceHd : selCeHd)
		{
			roundDlg->roundUtil2d.SetUpRoundConstEdge(*(const YsShellExt *)slHd,ceHd,NULL);
		}

		roundDlg->Make((const YsShell *)slHd);

		const double maxRadius=roundDlg->roundUtil2d.CalculateMaximumRadius(*(const YsShell *)slHd);
		if(YsTolerance<maxRadius)
		{
			Edit_ClearUIMode();

			roundDlg->mode3D=YSFALSE;
			roundDlg->maxRadius=maxRadius;

			const double radius=maxRadius*roundDlg->slider->GetPosition();
			roundDlg->roundUtil2d.CalculateRoundingAll(*(const YsShell *)slHd,radius,4);

			AddDialog(roundDlg);
			ArrangeDialog();

			undoRedoRequireToClearUIMode=YSTRUE;
			deletionRequireToClearUIMode=YSTRUE;

			UIDrawCallBack3D=FsGui3DMainCanvas::Edit_Round_DrawCallBack2d;
			SpaceKeyCallBack=FsGui3DMainCanvas::Edit_Round_SpaceKeyCallBack;

			return;
		}
	}
}



/*static*/ void FsGui3DMainCanvas::Edit_Round_PolygonOrConstEdgeSelectedCorner(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;
	canvas->Edit_Round_PolygonOrConstEdgeSelectedCorner(); // Test Implementation
}

void FsGui3DMainCanvas::Edit_Round_PolygonOrConstEdgeSelectedCorner(void)
{
	if(NULL!=slHd)
	{
		YsShellExtEdit::StopIncUndo incUndo(slHd);

		YsArray <YsShellVertexHandle> selVtHd;
		slHd->GetSelectedVertex(selVtHd);

		YsArray <YsShellPolygonHandle> selPlHd;
		slHd->GetSelectedPolygon(selPlHd);

		YsArray <YsShellExt::ConstEdgeHandle> selCeHd;
		slHd->GetSelectedConstEdge(selCeHd);

		YsShellVertexStore vtHdToRound(*(const YsShell *)slHd);
		for(auto vtHd : selVtHd)
		{
			vtHdToRound.AddVertex(vtHd);
		}

		roundDlg->roundUtil2d.CleanUp();

		for(auto plHd : selPlHd)
		{
			roundDlg->roundUtil2d.SetUpRoundPolygon(*(const YsShell *)slHd,plHd,&vtHdToRound);
		}

		for(auto ceHd : selCeHd)
		{
			roundDlg->roundUtil2d.SetUpRoundConstEdge(*(const YsShellExt *)slHd,ceHd,&vtHdToRound);
		}

		roundDlg->Make((const YsShell *)slHd);

		const double maxRadius=roundDlg->roundUtil2d.CalculateMaximumRadius(*(const YsShell *)slHd);
		if(YsTolerance<maxRadius)
		{
			Edit_ClearUIMode();

			roundDlg->mode3D=YSFALSE;
			roundDlg->maxRadius=maxRadius;

			const double radius=maxRadius*roundDlg->slider->GetPosition();
			roundDlg->roundUtil2d.CalculateRoundingAll(*(const YsShell *)slHd,radius,4);

			AddDialog(roundDlg);
			ArrangeDialog();

			undoRedoRequireToClearUIMode=YSTRUE;
			deletionRequireToClearUIMode=YSTRUE;

			UIDrawCallBack3D=FsGui3DMainCanvas::Edit_Round_DrawCallBack2d;
			SpaceKeyCallBack=FsGui3DMainCanvas::Edit_Round_SpaceKeyCallBack;

			return;
		}
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_Round_DrawCallBack2d(FsGui3DMainCanvas &canvas)
{
	canvas.Edit_Round_DrawCallBack2d();
}

void FsGui3DMainCanvas::Edit_Round_DrawCallBack2d(void)
{
	if(NULL!=slHd)
	{
		YsGLVertexBuffer vtxBuf;
		for(auto corner : roundDlg->roundUtil2d.cornerArray)
		{
			for(int side=0; side<2; ++side)
			{
				for(YSSIZE_T idx=0; idx<corner.subDiv[side].GetN()-1; ++idx)
				{
					vtxBuf.AddVertex(corner.subDiv[side][idx].pos);
					vtxBuf.AddVertex(corner.subDiv[side][idx+1].pos);
				}
				vtxBuf.AddVertex(corner.subDiv[side].Last().pos);
				vtxBuf.AddVertex(corner.roundedCornerPos);
			}
		}

		struct YsGLSL3DRenderer *renderer=YsGLSLSharedFlat3DRenderer();
		YsGLSLUse3DRenderer(renderer);

		const GLfloat edgeCol[4]={0,0,0,1.0f};
		YsGLSLSet3DRendererUniformColorfv(renderer,edgeCol);
		YsGLSLDrawPrimitiveVtxfv(renderer,GL_LINES,vtxBuf.GetNumVertex(),vtxBuf);

		YsGLSLEndUse3DRenderer(renderer);
	}}

YSRESULT FsGui3DMainCanvas::Edit_Round2d(void)
{
	const int nDiv=roundDlg->subdivTxt->GetInteger();
	const double radius=roundDlg->slider->GetPosition()*roundDlg->maxRadius;

	if(YSOK!=roundDlg->roundUtil2d.CalculateRoundingAll(*(const YsShell *)slHd,radius,nDiv))
	{
		return YSERR;
	}

	YsShellExtEdit::StopIncUndo stopIncUndo(slHd);

	for(auto plKey : roundDlg->roundUtil2d.targetPlKeyArray)
	{
		const YsShellPolygonHandle plHd=slHd->FindPolygon(plKey);
		if(NULL==plHd)
		{
			continue;
		}

		YsArray <YsShellVertexHandle> plVtHd;
		slHd->GetPolygon(plVtHd,plHd);

		YsArray <YsShellExt_RoundUtil::VertexPositionPair> rounded=roundDlg->roundUtil2d.MakeRoundedVertexSequence(*(const YsShell *)slHd,plVtHd,YSTRUE);

		plVtHd.CleanUp();
		for(YSSIZE_T idx=0; idx<rounded.GetN(); ++idx)
		{
			if(NULL!=rounded[idx].vtHd)
			{
				plVtHd.Append(rounded[idx].vtHd);
				slHd->SetVertexPosition(rounded[idx].vtHd,rounded[idx].pos);
			}
			else
			{
				YsShellVertexHandle vtHd=slHd->AddVertex(rounded[idx].pos);
				plVtHd.Append(vtHd);
			}
		}

		slHd->SetPolygonVertex(plHd,plVtHd.GetN(),plVtHd);
	}
	for(auto ceKey : roundDlg->roundUtil2d.targetCeKeyArray)
	{
		const YsShellExt::ConstEdgeHandle ceHd=slHd->FindConstEdge(ceKey);
		if(NULL==ceHd)
		{
			continue;
		}

		YsArray <YsShellVertexHandle> ceVtHd;
		YSBOOL isLoop;
		slHd->GetConstEdge(ceVtHd,isLoop,ceHd);

		YsArray <YsShellExt_RoundUtil::VertexPositionPair> rounded=roundDlg->roundUtil2d.MakeRoundedVertexSequence(*(const YsShell *)slHd,ceVtHd,isLoop);

		ceVtHd.CleanUp();
		for(YSSIZE_T idx=0; idx<rounded.GetN(); ++idx)
		{
			if(NULL!=rounded[idx].vtHd)
			{
				ceVtHd.Append(rounded[idx].vtHd);
				slHd->SetVertexPosition(rounded[idx].vtHd,rounded[idx].pos);
			}
			else
			{
				YsShellVertexHandle vtHd=slHd->AddVertex(rounded[idx].pos);
				ceVtHd.Append(vtHd);
			}
		}

		slHd->ModifyConstEdge(ceHd,ceVtHd,isLoop);
	}

	needRemakeDrawingBuffer|=(NEED_REMAKE_DRAWING_VERTEX|NEED_REMAKE_DRAWING_POLYGON|NEED_REMAKE_DRAWING_CONSTEDGE|
	                          NEED_REMAKE_DRAWING_SELECTED_VERTEX|
	                          NEED_REMAKE_DRAWING_SELECTED_POLYGON|NEED_REMAKE_DRAWING_SELECTED_CONSTEDGE);
	SetNeedRedraw(YSTRUE);

	return YSOK;
}

////////////////////////////////////////////////////////////

/*static*/ void FsGui3DMainCanvas::Edit_Round_AroundVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;

	if(NULL!=canvas->slHd)
	{
		YsArray <YsShellVertexHandle> selVtHd;
		canvas->slHd->GetSelectedVertex(selVtHd);

		canvas->roundDlg->Make((const YsShell *)canvas->slHd);
		if(YSOK==canvas->roundDlg->roundUtil3d.SetUpRoundVertex(*(const YsShell *)canvas->slHd,selVtHd))
		{
			const double maxRadius=canvas->roundDlg->roundUtil3d.CalculateMaximumRadius(*(const YsShell *)canvas->slHd);

			if(YsTolerance<maxRadius)
			{
				canvas->Edit_ClearUIMode();

				canvas->roundDlg->mode3D=YSTRUE;
				canvas->roundDlg->maxRadius=maxRadius;

				const double radius=maxRadius*canvas->roundDlg->slider->GetPosition();
				canvas->roundDlg->roundUtil3d.CalculateRoundingAll(*(const YsShell *)canvas->slHd,radius);
				canvas->roundDlg->roundUtil3d.SubdivideByNumberOfSubdivision(4);

				canvas->AddDialog(canvas->roundDlg);
				canvas->ArrangeDialog();

				canvas->undoRedoRequireToClearUIMode=YSTRUE;
				canvas->deletionRequireToClearUIMode=YSTRUE;

				canvas->UIDrawCallBack3D=FsGui3DMainCanvas::Edit_Round_DrawCallBack3d;
				canvas->SpaceKeyCallBack=FsGui3DMainCanvas::Edit_Round_SpaceKeyCallBack;

				return;
			}
		}
	}

	canvas->MessageDialog(FSGUI_COMMON_ERROR,FSGUI_ERROR_CANNOTROUNDVTX);
}

/*static*/ void FsGui3DMainCanvas::Edit_Round_AroundSelectedPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;

	YsArray <YsShellPolygonHandle> selPlHd;
	canvas->slHd->GetSelectedPolygon(selPlHd);

	canvas->roundDlg->Make((const YsShell *)canvas->slHd);
	if(YSOK==canvas->roundDlg->roundUtil3d.SetUpForAroundPolygonGroup(*(const YsShellExt *)canvas->slHd,selPlHd.GetN(),selPlHd))
	{
		const double maxRadius=canvas->roundDlg->roundUtil3d.CalculateMaximumRadius(*(const YsShell *)canvas->slHd);

		if(YsTolerance<maxRadius)
		{
			canvas->Edit_ClearUIMode();

			canvas->roundDlg->mode3D=YSTRUE;
			canvas->roundDlg->maxRadius=maxRadius;

			const double radius=maxRadius*canvas->roundDlg->slider->GetPosition();
			canvas->roundDlg->roundUtil3d.CalculateRoundingAll(*(const YsShell *)canvas->slHd,radius);
			canvas->roundDlg->roundUtil3d.SubdivideByNumberOfSubdivision(4);

			canvas->AddDialog(canvas->roundDlg);
			canvas->ArrangeDialog();

			canvas->undoRedoRequireToClearUIMode=YSTRUE;
			canvas->deletionRequireToClearUIMode=YSTRUE;

			canvas->UIDrawCallBack3D=FsGui3DMainCanvas::Edit_Round_DrawCallBack3d;
			canvas->SpaceKeyCallBack=FsGui3DMainCanvas::Edit_Round_SpaceKeyCallBack;

			return;
		}
	}

	canvas->MessageDialog(FSGUI_COMMON_ERROR,FSGUI_ERROR_CANNOTROUNDPLGBOUNDARY);
}

/*static*/ void FsGui3DMainCanvas::Edit_Round_OpenVertexSequence(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *)
{
	FsGui3DMainCanvas *canvas=(FsGui3DMainCanvas *)appPtr;

	auto selVtHd=canvas->slHd->GetSelectedVertex();
	auto selCeHd=canvas->slHd->GetSelectedConstEdge();

	if(0<selVtHd.GetN() && 0<selCeHd.GetN())
	{
		canvas->MessageDialog(FSGUI_COMMON_ERROR,FSGUI_ERROR_NEEDSELECTVTXORCONSTEDGE);
		return;
	}

	if(0<selCeHd.GetN())
	{
		selVtHd=YsShellExt_TrackingUtil::MakeVertexSequenceFromMultiConstEdge(*(const YsShellExt *)canvas->slHd,selCeHd);
		if(0==selVtHd.GetN())
		{
			canvas->MessageDialog(FSGUI_COMMON_ERROR,FSGUI_ERROR_CANNOTMAKECONSTEDGECHAIN);
			return;
		}
	}

	canvas->roundDlg->Make((const YsShell *)canvas->slHd);
	if(YSOK==canvas->roundDlg->roundUtil3d.SetUpForVertexSequence(*(const YsShellExt *)canvas->slHd,selVtHd))
	{
		const double maxRadius=canvas->roundDlg->roundUtil3d.CalculateMaximumRadius(*(const YsShell *)canvas->slHd);

		printf("maxRadius %lf\n",maxRadius);

		if(YsTolerance<maxRadius)
		{
			canvas->Edit_ClearUIMode();

			canvas->roundDlg->mode3D=YSTRUE;
			canvas->roundDlg->maxRadius=maxRadius;

			const double radius=maxRadius*canvas->roundDlg->slider->GetPosition();
			canvas->roundDlg->roundUtil3d.CalculateRoundingAll(*(const YsShell *)canvas->slHd,radius);
			canvas->roundDlg->roundUtil3d.SubdivideByNumberOfSubdivision(4);

			canvas->AddDialog(canvas->roundDlg);
			canvas->ArrangeDialog();

			canvas->undoRedoRequireToClearUIMode=YSTRUE;
			canvas->deletionRequireToClearUIMode=YSTRUE;

			canvas->UIDrawCallBack3D=FsGui3DMainCanvas::Edit_Round_DrawCallBack3d;
			canvas->SpaceKeyCallBack=FsGui3DMainCanvas::Edit_Round_SpaceKeyCallBack;
		}
		else
		{
			canvas->MessageDialog(FSGUI_COMMON_ERROR,FSGUI_ERROR_CANNOTROUNDPLGBOUNDARY);
		}
	}
}

/*static*/ void FsGui3DMainCanvas::Edit_Round_DrawCallBack3d(FsGui3DMainCanvas &canvas)
{
	canvas.Edit_Round_DrawCallBack3d();
}

void FsGui3DMainCanvas::Edit_Round_DrawCallBack3d(void)
{
	if(NULL!=slHd)
	{
		YsGLVertexBuffer vtxBuf;
		for(auto corner : roundDlg->roundUtil3d.cornerArray)
		{
			if(0<corner.subDiv.GetN())
			{
				for(YSSIZE_T idx=0; idx<corner.subDiv.GetN()-1; ++idx)
				{
					vtxBuf.AddVertex(corner.subDiv[idx].pos);
					vtxBuf.AddVertex(corner.subDiv[idx+1].pos);
				}
				vtxBuf.AddVertex(corner.subDiv.Last().pos);
				vtxBuf.AddVertex(corner.roundedCornerPos);
			}
		}

		struct YsGLSL3DRenderer *renderer=YsGLSLSharedFlat3DRenderer();
		YsGLSLUse3DRenderer(renderer);

		const GLfloat edgeCol[4]={0,0,0,1.0f};
		YsGLSLSet3DRendererUniformColorfv(renderer,edgeCol);
		YsGLSLDrawPrimitiveVtxfv(renderer,GL_LINES,vtxBuf.GetNumVertex(),vtxBuf);

		YsGLSLEndUse3DRenderer(renderer);
	}
}

/*static*/ YSRESULT FsGui3DMainCanvas::Edit_Round_SpaceKeyCallBack(FsGui3DMainCanvas &canvas)
{
	return canvas.Edit_Round_SpaceKeyCallBack();
}

YSRESULT FsGui3DMainCanvas::Edit_Round_SpaceKeyCallBack(void)
{
	if(YSTRUE==roundDlg->mode3D)
	{
		return Edit_Round3d();
	}
	else
	{
		return Edit_Round2d();
	}
}


YSRESULT FsGui3DMainCanvas::Edit_Round3d(void)
{
	const int nDiv=roundDlg->subdivTxt->GetInteger();
	const double radius=roundDlg->slider->GetPosition()*roundDlg->maxRadius;

	if(0>=nDiv || 256<nDiv)
	{
		MessageDialog(FSGUI_COMMON_ERROR,FSGUI_ERROR_DIVCOUNTOUTOF1AND256);
		return YSERR;
	}

	if(NULL!=slHd)
	{
		YsShellExtEdit &shl=*(YsShellExtEdit *)slHd;

		if(YSOK==YsShellExtEdit_RoundUtil3d::Apply(shl,roundDlg->roundUtil3d,radius,nDiv))
		{
			needRemakeDrawingBuffer|=(NEED_REMAKE_DRAWING_VERTEX|NEED_REMAKE_DRAWING_POLYGON|NEED_REMAKE_DRAWING_CONSTEDGE|
			                          NEED_REMAKE_DRAWING_SELECTED_VERTEX|
			                          NEED_REMAKE_DRAWING_SELECTED_POLYGON|NEED_REMAKE_DRAWING_SELECTED_CONSTEDGE);
			SetNeedRedraw(YSTRUE);

			return YSOK;
		}
	}
	return YSERR;
}

