/* ////////////////////////////////////////////////////////////

  File Name: fsgui3dapp.h
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#ifndef FSGUI_DAPP_IS_INCLUDED
#define FSGUI_DAPP_IS_INCLUDED
/* { */

#include <fsgui3d.h>

#include <ysshellkernel.h>
#include <ysglbuffer.h>

#include <ysshellextedit_duplicateutil.h>

// Call back functions from the main framework. >>
YSBOOL FsGui3dAppMustTerminate(void);
int FsGui3dAppSleepIntervalInMS(void);

void FsGui3DAppInitialize(int ac,char *av[]);
void FsGui3DAppIntervalCallBack(void);
void FsGui3dAppCleanUp(void);
// Call back functions from the main framework. <<

class FsGui3DMainMenu : public FsGuiPopUpMenu
{
private:
	class FsGui3DMainCanvas *canvas;
public:
	FsGui3DMainMenu(class FsGui3DMainCanvas *canvas);
	virtual void OnSelectMenuItem(FsGuiPopUpMenuItem *item);

	class FsGui3DMainCanvas *GetMainCanvas(void) const;
};

class FsGui3DMainCanvas : public FsGuiCanvas
{
friend class PolyCreProjectionDialog;
friend class PolyCreEditSweepConstEdgeDialog;
friend class PolyCreSlideShowDialog;
friend class PolyCreDnmTreeDialog;
friend class PolyCreGuiExtension;

public:
	class ShortCutKey
	{
	public:
		void (*callBackFunc)(void *,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
		int fskey;
		YSBOOL ctrl,shift;
	};

	enum
	{
		MENU_FILE_EXIT
	};

	YsWString lastAccessedFileName;

	class PolyCreFileName *fileAndDirName;
	class PolyCreConfig *config;
	class PolyCreReferenceBitmapStore *refBmpStore;

	FsGui3DInterface threeDInterface;
	FsGui3DViewControl drawEnv;
	FsGui3DViewControlDialog *viewControlDlg;
	FsGui3DMainMenu *mainMenu;
	FsGuiStatusBar *topBar,*bottomBar;


	FsGuiPopUpMenu *fileRecent;
	FsGuiRecentFiles recent;

	FsGuiPopUpMenuItem *viewDrawAxis;
	FsGuiPopUpMenuItem *viewDrawVertex;
	FsGuiPopUpMenuItem *viewDrawConstEdge;
	FsGuiPopUpMenuItem *viewDrawPolygonFill;
	FsGuiPopUpMenuItem *viewDrawWireframe;
	FsGuiPopUpMenuItem *viewDrawShrunkPolygon;
	FsGuiPopUpMenuItem *viewDrawBackFaceInDifferentColor;
	FsGuiPopUpMenuItem *viewEnableViewPort;
	FsGuiPopUpMenuItem *viewUseSameViewPortForAll;
	FsGuiPopUpMenuItem *viewHighlightNonManifoldEdge;

	YsShellExtEdit_DuplicateUtil cutBuffer;


	// [Modeless dialogs]
	// Adding a modeless dialog:
	//   (1) Add a pointer in the following chunk.
	//   (2) Add an initialization in the constructor of the aplication.
	//   (3) Add deletion in the destructor of the application.
	//   (4) Add RemoveDialog in Edit_ClearUIIMode
	class PolyCreSolidOfRevolutionDialog *solidOfRevolutionDlg;
	class PolyCreDnmTreeDialog *dnmTreeDlg;
	class PolyCreAngleSliderDialog *angleSliderDlg;
	class PolyCreEquilateralDialog *equilateralDlg;
	class PolyCreSlideShowDialog *slideShowDlg;
	class PolyCreParallelogramDialog *parallelogramDlg;
	class PolyCreInsertVertexDialog *insertVtxDlg;
	class PolyCreChopOffDialog *chopOffDlg;
	class PolyCreExtendOrShrinkDialog *extendOrShrinkDlg;
	class YsColorPaletteDialog *colorPaletteDlg;
	class FsGuiFileDialog *fdlg;
	FsGuiMessageBoxDialog *mbox;
	FsGuiNumberDialog *numberDlg;
	class YsCreateAirfoilDialog *createAirfoilDlg;
	class PolyCreConfigDialog *configDlg;
	class PolyCreRefBmpDialog *refBmpDlg;
	class PolyCreEditParallelSweepDialog *parallelSweepDlg;
	class PolyCreEditParallelSweepWithPathDialog *parallelSweepWithPathDlg;
	class PolyCreEditSweepConstEdgeDialog *sweepConstEdgeDlg;
	class PolyCreEditCreateMirrorImageDialog *mirrorImageDlg;
	class PolyCreHemmingDialog *hemmingDlg;
	class PolyCreRoundDialog *roundDlg;
	class PolyCreCreatePrimitiveDialog *createPrimitiveDlg;
	class PolyCreProjectionDialog *projectionDlg;
	class PolyCreBoolDialog *boolDlg;

	// [Modal dialogs]
	class PolyCreAboutDialog *aboutDlg;
	class PolyCreHelpDialog *helpDlg;


	class YsShellExt_SweepInfo *sweepInfo;

	YSBOOLEANOPERATION boolOpType;

	YSBOOL undoRedoRequireToClearUIMode;
	YSBOOL deletionRequireToClearUIMode;

	YsArray <ShortCutKey> shortCut;

	class PolyCreGuiExtension *guiExtension;


	FsGui3DMainCanvas();
	~FsGui3DMainCanvas();
private:
	class DrawingCacheState
	{
	public:
		const YsShellDnmContainer <YsShellExtEdit>::Node *slHd;
		YsShellExtEdit::SavePoint whenCached;
		void Save(const YsShellDnmContainer <YsShellExtEdit>::Node *shl);
		YSBOOL IsModified(const YsShellDnmContainer <YsShellExtEdit>::Node *shl) const;
	};
	class NonManifoldEdgeCache : public DrawingCacheState
	{
	public:
		YsGLVertexBuffer lineVtxBuf;
		YsGLColorBuffer lineColBuf;

		YsGLVertexBuffer pointVtxBuf; // Stupid OpenGL ES2.0 got rid of glPointSize.  All points need to be drawn as triangles.  What a waste.
		YsGLPixOffsetBuffer pointPixOffsetBuf;
		YsGLColorBuffer pointColBuf;

		void CacheNonManifoldEdge(const YsShellDnmContainer <YsShellExtEdit>::Node *shl);
	};

	void AddViewControlDialog(void);
	void MakeMainMenu(void);
	void SetMenuCheck(void);
	void DeleteMainMenu(void);

public:
	void MessageDialog(const wchar_t title[],const wchar_t msg[]);
	void MessageDialog(const wchar_t title[],const wchar_t msg[],const wchar_t copyBtn[]);
	void YesNoDialog(
	    const wchar_t title[],const wchar_t msg[],const wchar_t yes[],const wchar_t no[],
	    void (*closeModalCallBackFunc)(FsGuiControlBase *,FsGuiDialog *,int));

	virtual void OnInterval(void);
	virtual void OnSelectMenuItem(FsGuiPopUpMenuItem *menuItem);
	virtual YSRESULT OnLButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	virtual YSRESULT OnLButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	virtual YSRESULT OnRButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	virtual YSRESULT OnRButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	virtual YSRESULT OnMouseMove(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my,int prevMx,int prevMy);

	virtual YSRESULT OnSaveView(const class FsGui3DViewControlDialog *dlg);
	static void OnSaveView_FileSelected(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);
	static void OnSaveView_ConfirmOverwriteCallBack(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);
	static void OnSaveView_OverwriteChecked(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);

	virtual YSRESULT OnLoadView(class FsGui3DViewControlDialog *dlg);
	static void OnLoadView_FileSelected(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);


	YSRESULT (*ModeCleanUpCallBack)(FsGui3DMainCanvas &canvas);
	YSRESULT (*SpaceKeyCallBack)(FsGui3DMainCanvas &canvas);
	YSRESULT (*BackSpaceKeyCallBack)(FsGui3DMainCanvas &canvas);
	YSRESULT (*LButtonCallBack)(FsGui3DMainCanvas &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSBOOL lBtnDown;  // Shouldn't be turned on if lButtonDown is caught by other UI widgets.
	int lBtnDownMx,lBtnDownMy;
	int moveMx,moveMy;
	YSRESULT (*RButtonCallBack)(FsGui3DMainCanvas &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT (*LButtonBoxCallBack)(FsGui3DMainCanvas &canvas,int xMin,int yMin,int xMax,int yMax);

	YSRESULT OnKeyDown(int fskey,YSBOOL shift,YSBOOL ctrl,YSBOOL alt);
	void Draw(void);
private:
	void DrawInGeometryEditMode(void);
	void DrawInDnmEditMode(void);
	void DrawInDnmEditModeRecursive(YsShellDnmContainer <YsShellExtEdit>::Node *,const YsShellDnmContainer <YsShellExtEdit>::DnmState &state,const YsMatrix4x4 &curTfm);

public:
	void RemakeDrawingBuffer(
	    YsShellDnmContainer <YsShellExtEdit>::Node *slHd,
	    unsigned int needRemakeDrawingBuffer);
	void AddPolygonFragment(YsShellPolygonHandle plHd,const YsVec3 &nom,const YsShellVertexHandle triVtHd[3],YSBOOL flatShaded);

	YsShellVertexHandle PickedVtHd(int mx,int my,YSSIZE_T nExclude,const YsShellVertexHandle excludeVtHd[]) const;
	template <const int N>
	YsShellVertexHandle PickedVtHd(int mx,int my,YsArray <YsShellVertexHandle,N> &excludeVtHd) const;

	YsShellVertexHandle PickedVtHdAmongSelection(int mx,int my) const;	// Selected vertices are cached in the vertex buffer.  Can be calculated faster.

	YsArray <YsPair <YsShellVertexHandle,double> > BoxedVtHd(int x0,int y0,int x1,int y1,YSSIZE_T nExclude,const YsShellVertexHandle excludeVtHd[]) const;
	template <const int N>
	YsArray <YsPair <YsShellVertexHandle,double> > BoxedVtHd(int x0,int y0,int x1,int y1,const YsArray <YsShellVertexHandle,N> &excludeVtHd) const;

	YsArray <YsPair <YsShellVertexHandle,double> > BoxedVtHdAmongSelection(int x0,int y0,int x1,int y1) const;

	YsShellPolygonHandle PickedPlHd(int mx,int my,YSSIZE_T nExclude,const YsShellPolygonHandle excludePlHd[]) const;
	template <const int N>
	YsShellPolygonHandle PickedPlHd(int mx,int my,YsArray <YsShellPolygonHandle,N> &excludePlHd) const;

	YsShellPolygonHandle PickedPlHdAmongSelection(int mx,int my) const;

	YsArray <YsPair <YsShellPolygonHandle,double> > BoxedPlHd(int x0,int y0,int x1,int y1,YSSIZE_T nExclude,const YsShellPolygonHandle excludePlHd[]) const;
	template <const int N>
	YsArray <YsPair <YsShellPolygonHandle,double> > BoxedPlHd(int x0,int y0,int x1,int y1,YsArray <YsShellPolygonHandle,N> &excludePlHd) const;

	YsArray <YsPair <YsShellPolygonHandle,double> > BoxedPlHdAmongSelection(int x0,int y0,int x1,int y1) const;

	YsShellExt::ConstEdgeHandle PickedCeHd(int mx,int my,YSSIZE_T nExclude,const YsShellExt::ConstEdgeHandle excludeCeHd[]) const;
	template <const int N>
	YsShellExt::ConstEdgeHandle PickedCeHd(int mx,int my,YsArray <YsShellExt::ConstEdgeHandle,N> &excludeCeHd) const;

	YsShellExt::ConstEdgeHandle PickedCeHdAmongSelection(int mx,int my) const;

	YsArray <YsPair <YsShellExt::ConstEdgeHandle,double> > BoxedCeHd(int x0,int y0,int x1,int y1,YSSIZE_T nExclude,const YsShellExt::ConstEdgeHandle excludeCeHd[]) const;
	template <const int N>
	YsArray <YsPair <YsShellExt::ConstEdgeHandle,double> > BoxedCeHd(int x0,int y0,int x1,int y1,const YsArray <YsShellExt::ConstEdgeHandle,N> &excludeCeHd) const;

	YsArray <YsPair <YsShellExt::ConstEdgeHandle,double> > BoxedCeHdAmongSelection(int x0,int y0,int x1,int y1) const;

	YsShellExt::FaceGroupHandle PickedFgHd(int mx,int my,int nExclude,const YsShellExt::FaceGroupHandle excludeFgHd[]) const;
	template <const int N>
	YsShellExt::FaceGroupHandle PickedFgHd(int mx,int my,YsArray <YsShellExt::FaceGroupHandle,N> &excludeFgHd) const;

	YsShellExt::FaceGroupHandle PickedFgHdAmongSelection(int mx,int my) const;

	const YsShellDnmContainer <YsShellExtEdit>::Node *PickedSlHd(int mx,int my,YSSIZE_T nExclude,const YsShellDnmContainer <YsShellExtEdit>::Node *excludeSlHd[]) const;
	template <const int N>
	const YsShellDnmContainer <YsShellExtEdit>::Node *PickedSlHd(int mx,int my,YsArray <const YsShellDnmContainer <YsShellExtEdit>::Node *> &excludeSlHd) const;

	YsShellDnmContainer <YsShellExtEdit>::Node *PickedSlHd(int mx,int my,YSSIZE_T nExclude,const YsShellDnmContainer <YsShellExtEdit>::Node *excludeSlHd[]);
	template <const int N>
	YsShellDnmContainer <YsShellExtEdit>::Node *PickedSlHd(int mx,int my,YsArray <const YsShellDnmContainer <YsShellExtEdit>::Node *> &excludeSlHd);

private:
	int PickedIndexInPointBuffer(double &z,int mx,int my,const YsGLVertexBuffer &pointVtxBuffer,const YsArray <int> &pointIdxBuffer,const YsKeyStore &exclusion) const;
	int PickedIndexInEdgeBuffer(double &z,int mx,int my,const YsGLVertexBuffer &edgeVtxBuffer,const YsArray <int> &edgeIdxBuffer,const YsKeyStore &exclusion) const;
	int PickedIndexInTriangleBuffer(double &z,int mx,int my,const YsGLVertexBuffer &triVtxBuffer,const YsArray <int> &triIdxBuffer,const YsKeyStore *limit,const YsKeyStore *exclusion) const;

	YsArray <YsPair <int,double> > BoxedIndexInPointBuffer(int x0,int y0,int x1,int y1,const YsGLVertexBuffer &pointVtxBuffer,const YsArray <int> &pointIdxBuffer,const YsKeyStore &exclusion) const;
	YsArray <YsPair <int,double> > BoxedIndexInEdgeBuffer(
	    int x0,int y0,int x1,int y1,const YsGLVertexBuffer &edgeVtxBuffer,const YsArray <int> &edgeIdxBuffer,const YsKeyStore *limit,const YsKeyStore *exclusion) const;
	YsArray <YsPair <int,double> > BoxedIndexInTriangleBuffer(
	    int x0,int y0,int x1,int y1,const YsGLVertexBuffer &triVtxBuffer,const YsArray <int> &triIdxBuffer,const YsKeyStore *limit,const YsKeyStore *exclusion) const;

public:
	enum
	{
		NEED_REMAKE_DRAWING_NONE=0,
		NEED_REMAKE_DRAWING_ALL=~(unsigned int)0,
		NEED_REMAKE_DRAWING_VERTEX=1,
		NEED_REMAKE_DRAWING_SELECTED_VERTEX=2,
		NEED_REMAKE_DRAWING_POLYGON=4,
		NEED_REMAKE_DRAWING_SELECTED_POLYGON=8,
		NEED_REMAKE_DRAWING_CONSTEDGE=16,
		NEED_REMAKE_DRAWING_SELECTED_CONSTEDGE=32,
		NEED_REMAKE_DRAWING_SELECTED_FACEGROUP=64,
	};
	GLfloat vertexPickRadius;
	unsigned int needRemakeDrawingBuffer;
	unsigned int needRemakeDrawingBuffer_for_OtherShell;
	YSBOOL drawAxis;
	YSBOOL drawVertex,actuallyDrawVertex;
	YSBOOL drawPolygonEdge,actuallyDrawPolygonEdge;
	YSBOOL drawPolygon,actuallyDrawPolygon;
	YSBOOL drawConstEdge,actuallyDrawConstEdge;
	YSBOOL drawShrunkPolygon,actuallyDrawShrunkPolygon;
	YSBOOL drawOtherShell,actuallyDrawOtherShell;
	YSBOOL drawBackFaceInDifferentColor;
	YSBOOL drawApplyViewPortToAllShell;
	YsColor backFaceColor;

	int dnmCurrentState; // -1:Static mode
	YSBOOL dnmEditMode;
	YsShellDnmContainer <YsShellExtEdit>::DnmState dnmState;

	GLfloat normalVertexMarkerRadius;
	GLfloat roundVertexMarkerRadius;
	GLfloat selectedVertexMarkerRadius;
	GLfloat polygonShrinkRatioForPicking;

private:
	YSBOOL drawNonManifoldEdge,actuallyDrawNonManifoldEdge;
	mutable NonManifoldEdgeCache nonManifoldEdgeCache;
	

	void (*UIBeforeDrawCallBack)(FsGui3DMainCanvas &canvas);
	void (*UIDrawCallBack2D)(FsGui3DMainCanvas &canvas);
	void (*UIDrawCallBack3D)(FsGui3DMainCanvas &canvas);
	YsGLVertexBuffer UILineVtxBuf;

private:
	YsShellDnmContainer <YsShellExtEdit> shlGrp;
	YsShellDnmContainer <YsShellExtEdit>::Node *slHd;

public:
	YsShellExtEdit *Slhd(void);
	void MakeInitialEmptyShell(void);
	void LoadGeneral(const wchar_t fn[],YsShellDnmContainer <YsShellExtEdit>::Node *parent);
	void SaveGeneral(const YsShellExtEdit &shl,const wchar_t fn[]);
	YsWString GetDataDir(void) const;
	void AddShortCut(int fskey,YSBOOL shift,YSBOOL ctrl,void (*callBackFunc)(void *,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *));


	void RefreshRecentlyUsedFileList(void);
	void AddRecentlyUsedFile(const wchar_t wfn[]);


	// Menu Call Back Functions >>
	static void File_New(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void File_Open(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void File_Open_FileSelected(FsGuiControlBase *canvas,FsGuiDialog *,int);
	static void File_OpenDnm_ConfirmCloseCallBack(FsGuiControlBase *canvas,FsGuiDialog *,int);
	void File_Open_AfterOpenFile(YsShellDnmContainer <YsShellExtEdit>::Node *newSlHd,const wchar_t fn[]);

	static void File_Save(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void File_SaveAs(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void File_SaveAs_FileSelected(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);
	static void File_SaveAs_STLChecked(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);
	void File_SaveAs_CheckOverwrite(void);
	static void File_SaveAs_ConfirmOverwriteCallBack(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);
	void File_SaveAs_PassedOverWriteCheck(const wchar_t fn[]);

	static void File_Close(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void File_Close_ConfirmCloseCallBack(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);
	void File_Close_Confirmed(void);

	static void File_Exit(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void File_Exit_ConfirmExitCallBack(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);

	static void File_SlideShow(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void File_SlideShow_FileSelected(FsGuiControlBase *canvas,FsGuiDialog *,int);
	static YSRESULT File_SlideShow_ModeCleanUpCallBack(FsGui3DMainCanvas &canvas);

	static void File_GetInfo_Statistics(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void File_GetInfo_Volume(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void File_Recent(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);



	static void Edit_ClearUIMode(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	void Edit_ClearUIMode(void);
	void Edit_ClearUIModeExceptDNMEditMode(void);

	static void Edit_Undo(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void Edit_Redo(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);

	static void Edit_Copy(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void Edit_Paste(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void Edit_PasteSpecial(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);

	static void Edit_DeleteSelection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void Edit_DeleteSelection(void);

	static void Edit_ForceDeleteSelection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_InsertVertex(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static YSRESULT Edit_InsertVertex_SpaceKeyCallBack(FsGui3DMainCanvas &canvas);

	static void Edit_InsertPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_InsertConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_InsertFaceGroup(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_SmoothShadeSelectedVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_FlatShadeSelectedVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Paint_ByPickingPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Select_Paint_ByPickingPolygon_LButtonCallBack(FsGui3DMainCanvas &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

	static void Edit_Paint_ByPickingFaceGroup(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Paint_SelectedPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Paint_All(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Paint_FloodFill_SameColor(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Edit_Paint_FloodFill_SameColor_LButtonCallBack(FsGui3DMainCanvas &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

	static void Edit_Paint_FloodFill_BoundedByConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Edit_Paint_FloodFill_BoundedByConstEdge_LButtonCallBack(FsGui3DMainCanvas &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);


	static void Edit_CreatePrimitiveShape(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Edit_CreatePrimitive_SpaceKeyCallBack(FsGui3DMainCanvas &canvas);
	static void Edit_CreatePrimitive_CursorMoved(void *appPtr,FsGui3DInterface &);
	static void Edit_CreatePrimitive_UIDrawCallBack3D(FsGui3DMainCanvas &canvas);


	static void Edit_CreateEquilateral(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_CreateEquilateral_DrawCallBack3D(FsGui3DMainCanvas &canvas);
	static void Edit_CreateEquilateral_3DInterfaceCallBack(void *appPtr,class FsGui3DInterface &itfc);
	static YSRESULT Edit_CreateEquilateral_SpaceKeyCallBack(FsGui3DMainCanvas &canvas);
	YSRESULT Edit_CreateEquilateral_SpaceKeyCallBack(void);


	static void Edit_CreateAirfoil(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_CreateAirfoil_DrawCallBack3D(FsGui3DMainCanvas &canvas);
	void Edit_CreateAirfoil_DrawCallBack3D(void);
	static void Edit_CreateAirfoil_PivotMoved(void *appPtr,FsGui3DInterface &itfc);
	void Edit_CreateAirfoil_Create(void);

	static void Edit_CreateMirrorImage(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void Edit_CreateMirrorImage_Create(void);

	static void Edit_CreateParallelogramFromTwoVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_CreateParallelogramFromTwoVertex_DrawCallBack3D(FsGui3DMainCanvas &canvas);
	static void Edit_CreateParallelogramFromTwoVertex_3DInterfaceCallBack(void *appPtr,class FsGui3DInterface &itfc);
	static YSRESULT Edit_CreateParallelogramFromTwoVertex_SpaceKeyCallBack(FsGui3DMainCanvas &canvas);
	YSRESULT Edit_CreateParallelogramFromTwoVertex_SpaceKeyCallBack(void);

	static void Edit_CreateParallelogramFromThreeVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_CreateSolidOfRevolution(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_CreateSolidOfRevolution_DrawCallBack3D(FsGui3DMainCanvas &canvas);
	static void Edit_CreateSolidOfRevolution_3DInterfaceCallBack(void *appPtr,class FsGui3DInterface &itfc);
	static YSRESULT Edit_CreateSolidOfRevolution_SpaceKeyCallBack(FsGui3DMainCanvas &canvas);
	YSRESULT Edit_CreateSolidOfRevolution_SpaceKeyCallBack(void);



	static void Edit_MoveSelection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_MoveSelection_DrawCallBack3D(FsGui3DMainCanvas &canvas);
	static void Edit_MoveSelection_CursorMoved(void *appPtr,FsGui3DInterface &itfc);
	static YSRESULT Edit_MoveSelection_SpaceKeyCallBack(FsGui3DMainCanvas &canvas);

	static void Edit_SwapEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_MergePolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_SplitPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_VertexCollapseMidPoint(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_VertexCollapseAtoB(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void Edit_VertexCollapse(YsShellVertexHandle vtHdToDel,YsShellVertexHandle vtHdToSurvive,const YsVec3 &newPos);

	static void Edit_Hemming_AroundSelectedPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_Hemming_AroundSelectedPolygon_UIDrawCallBack3D(FsGui3DMainCanvas &canvas);

	static void Edit_Hemming_AroundOutsideOfSelectedPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Hemming_OneSideOfVertexSequence(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	void Edit_Hemming_AroundSelectedPolygon(void);

	static void Edit_Intersection_CutByPlane(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Intersection_ChopOff(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_Intersection_ChopOff_DrawCallBack3D(FsGui3DMainCanvas &canvas);
	static void Edit_Intersection_ChopOff_3DInterfaceCallBack(void *appPtr,class FsGui3DInterface &itfc);
	static YSRESULT Edit_Intersection_ChopOff_SpaceKeyCallBack(FsGui3DMainCanvas &canvas);
	YSRESULT Edit_Intersection_ChopOff_SpaceKeyCallBack(void);

	static void Edit_Sweep_ParallelSweep(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_Sweep_ParallelSweep_DrawCallBack3D(FsGui3DMainCanvas &canvas);
	void Edit_Sweep_ParallelSweep_DrawCallBack3D(void);
	static YSRESULT Edit_Sweep_ParallelSweep_SpaceKeyCallBack(FsGui3DMainCanvas &canvas);
	YSRESULT Edit_Sweep_ParallelSweep_SpaceKeyCallBack(void);

	static void Edit_Sweep_ParallelSweepWithPath(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_Sweep_ParallelSweepWithPath_DrawCallBack3D(FsGui3DMainCanvas &canvas);
	void Edit_Sweep_ParallelSweepWithPath_DrawCallBack3D(void);
	static YSRESULT Edit_Sweep_ParallelSweepWithPath_SpaceKeyCallBack(FsGui3DMainCanvas &canvas);
	YSRESULT Edit_Sweep_ParallelSweepWithPath_SpaceKeyCallBack(void);

	static void Edit_Sweep_BetweenConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_Sweep_BetweenConstEdge_DrawCallBack3D(FsGui3DMainCanvas &canvas);
	static YSRESULT Edit_Sweep_BetweenConstEdge_SpaceKeyCallBack(FsGui3DMainCanvas &canvas);
	YSRESULT Edit_Sweep_BetweenConstEdge_Sweep(void);

	static void Edit_Sweep_ExtendOrShrink(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_Sweep_ExtendOrShrink_DrawCallBack3D(FsGui3DMainCanvas &canvas);
	static void Edit_Sweep_ExtendOrShrink_3DInterfaceCallBack(void *appPtr,class FsGui3DInterface &itfc);
	static YSRESULT Edit_Sweep_ExtendOrShrink_SpaceKeyCallBack(FsGui3DMainCanvas &canvas);
	YSRESULT Edit_Sweep_ExtendOrShrink_SpaceKeyCallBack(void);

	static void Edit_Round_PolygonOrConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void Edit_Round_PolygonOrConstEdge(void);

	static void Edit_Round_PolygonOrConstEdgeSelectedCorner(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void Edit_Round_PolygonOrConstEdgeSelectedCorner(void);

	static void Edit_Round_DrawCallBack2d(FsGui3DMainCanvas &canvas);
	void Edit_Round_DrawCallBack2d(void);

	YSRESULT Edit_Round2d(void);

	static void Edit_Round_AroundVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void Edit_Round_AroundVertex(const double radius,const int nDiv);

	static void Edit_Round_AroundSelectedPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Round_OpenVertexSequence(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Round_DrawCallBack3d(FsGui3DMainCanvas &canvas);
	void Edit_Round_DrawCallBack3d(void);
	static YSRESULT Edit_Round_SpaceKeyCallBack(FsGui3DMainCanvas &canvas);
	YSRESULT Edit_Round_SpaceKeyCallBack(void);

	YSRESULT Edit_Round3d(void);


	static void Edit_Projection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_Projection_DrawCallBack3d(FsGui3DMainCanvas &canvas);
	static YSRESULT Edit_Projection_SpaceKeyCallBack(FsGui3DMainCanvas &canvas);
	YSRESULT Edit_Projection_Project(void);

	static void Edit_TriangulateSelection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_InvertConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_SplitConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_MergeConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);



	static void BoolUnion(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void BoolSubtraction(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void BoolIntersection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void BoolMerge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void BoolSeparate(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Bool_LButtonCallBack(FsGui3DMainCanvas &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	void Bool_DiagnoseDialog(void);
	static void Bool_DiagnoseDialogCallBack(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);
	YSRESULT Bool_Apply(YSBOOLEANOPERATION boolOpType,YsShellExtEdit &shlA,YsShellExtEdit &shlB);

	static void Select_UnselectAll(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_PickVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Select_PickVertex_LButtonCallBack(FsGui3DMainCanvas &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	static YSRESULT Select_PickVertex_LButtonBoxCallBack(FsGui3DMainCanvas &canvas,int xMin,int yMin,int xMax,int yMax);
	static YSRESULT Select_PickVertex_BackSpaceKeyCallBack(FsGui3DMainCanvas &canvas);

	static void Select_PickPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Select_PickPolygon_LButtonCallBack(FsGui3DMainCanvas &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	static YSRESULT Select_PickPolygon_LButtonBoxCallBack(FsGui3DMainCanvas &canvas,int xMin,int yMin,int xMax,int yMax);
	static YSRESULT Select_PickPolygon_BackSpaceKeyCallBack(FsGui3DMainCanvas &canvas);

	static void Select_PickConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Select_PickConstEdge_LButtonCallBack(FsGui3DMainCanvas &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	static YSRESULT Select_PickConstEdge_LButtonBoxCallBack(FsGui3DMainCanvas &canvas,int xMin,int yMin,int xMax,int yMax);
	static YSRESULT Select_PickConstEdge_BackSpaceKeyCallBack(FsGui3DMainCanvas &canvas);

	static void Select_PickFaceGroup(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Select_PickFaceGroup_LButtonCallBack(FsGui3DMainCanvas &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	static YSRESULT Select_PickFaceGroup_BackSpaceKeyCallBack(FsGui3DMainCanvas &canvas);


	static void Select_UnselectVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_AllVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_InvertVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_VertexOfPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_VertexOfConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_VertexAlongBoundaryOrConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);


	static void Select_UnselectPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_AllPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_InvertPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_PolygonOfFaceGroup(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_PolygonConnected(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_PolygonUsingAtLeastOneOfSelectedVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);


	static void Select_CurrentShellByPicking(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Select_CurrentShellByPicking_LButtonCallBack(FsGui3DMainCanvas &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

	static void Select_NextShell(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_PrevShell(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);


	static void View_DrawAxis(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_DrawVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_DrawConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_DrawPolygonFill(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_DrawWireframe(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_DrawShrunkPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_DrawBackFaceInDifferentColor(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_FocusOn_SelectionCenterOfGravity(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_FocusOn_SelectionBoundingBoxCenter(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_FocusOn_Vertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_FocusOn_Vertex_RKey(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT View_FocusOn_Vertex_LButtonCallBack(FsGui3DMainCanvas &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT View_FocusOn_Vertex_Click(int mx,int my);

	static void View_FocusOn_Polygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_FocusOn_Polygon_ShiftRKey(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT View_FocusOn_Polygon_LButtonCallBack(FsGui3DMainCanvas &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

	static void View_FocusOn_PolygonStraight(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_FocusOn_Polygon_ShiftCtrlRKey(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT View_FocusOn_PolygonStraight_LButtonCallBack(FsGui3DMainCanvas &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

	YSRESULT View_FocusOn_Polygon_Click(int mx,int my,YSBOOL alsoChangeOrientation);

	void View_ApplyViewPortToAllShell(void);

	static void View_UseSameViewPortForAllShell(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_EnableViewPort(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_ResetViewPort(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_MoveViewPort(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_MoveViewPort_BoxMoved(void *appPtr,FsGui3DInterface &itfc);

	static void View_MoveViewPortToRotationCenter(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_ExpandViewPort(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_ShrinkViewPort(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_ViewPortFromSelection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_OpenRefBmpDialog(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_OpenRefBmpDialog_UIDrawCallBack3D(FsGui3DMainCanvas &canvas);

	static void View_LoadRefBmpConfig(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_LoadRefBmpConfig_FileSelected(FsGuiControlBase *canvas,FsGuiDialog *,int);

	static void View_SaveRefBmpConfig(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_SaveRefBmpConfig_FileSelected(FsGuiControlBase *canvas,FsGuiDialog *,int);
	static void View_SaveRefBmpConfig_PassedOverwriteCheck(FsGuiControlBase *canvas,FsGuiDialog *,int);
	void View_SaveRefBmpConfig_PassedOverwriteCheck(const wchar_t fn[]);

	static void View_ShowAllRefBmp(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_HideAllRefBmp(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_HighlightNonManifoldEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);



	static void Global_DeleteUnusedVertexAll(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Global_InvertPolygonAll(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Global_SetZeroNormalAll(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Global_FixOrientationAndNormalClosedVolume(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Global_TriangulateAll(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Global_DeleteRedundantVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Global_DeleteRedundantVertex_ToleranceEntered(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);

	static void Global_DeleteRedundantVertexNonManifoldEdgeOnly(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Global_DeleteRedundantVertexNonManifoldEdgeOnly_ToleranceEntered(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);

	void Global_DeleteRedundantVertex(const double tol,const YsShellVertexStore *limitedVtx);


	static void UtilMenu_ClearConstEdgeAll(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void UtilMenu_AddConstEdgeByDihedralAngle(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void UtilMenu_AddConstEdgeByDihedralAngle_DrawCallBack3D(FsGui3DMainCanvas &canvas);
	static YSRESULT UtilMenu_AddConstEdgeByDihedralAngle_SpaceKeyCallBack(FsGui3DMainCanvas &canvas);
	YSRESULT UtilMenu_AddConstEdgeByDihedralAngle_SpaceKeyCallBack(void);

	static void UtilMenu_AddConstEdgeAlongNonManifoldEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void UtilMenu_ClearFaceGroupAll(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void UtilMenu_ConstrainAllEdges(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void UtilMenu_MakeFaceGroupFromConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void UtilMenu_PaintBasedOnFaceGroup(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);




	static void Dnm_LoadStaticMode(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Dnm_LoadState0(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Dnm_LoadState1(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Dnm_LoadState2(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Dnm_LoadState3(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);



	static void Option_Config(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void Option_ReapplyConfig(class PolyCreConfig &cfg);
	void Option_ReapplyConfig(void);



	static void Help_About(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Help_Help(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	// Menu Call Back Functions <<
};

extern void FsGui3DSetAppMustTerminate(YSBOOL sw);

template <const int N>
YsShellVertexHandle FsGui3DMainCanvas::PickedVtHd(int mx,int my,YsArray <YsShellVertexHandle,N> &excludeVtHd) const
{
	return PickedVtHd(mx,my,excludeVtHd.GetN(),excludeVtHd);
}

template <const int N>
YsArray <YsPair <YsShellVertexHandle,double> > FsGui3DMainCanvas::BoxedVtHd(int x0,int y0,int x1,int y1,const YsArray <YsShellVertexHandle,N> &excludeVtHd) const
{
	return BoxedVtHd(x0,y0,x1,y1,excludeVtHd.GetN(),excludeVtHd);
}

template <const int N>
YsShellPolygonHandle FsGui3DMainCanvas::PickedPlHd(int mx,int my,YsArray <YsShellPolygonHandle,N> &excludePlHd) const
{
	return PickedPlHd(mx,my,excludePlHd.GetN(),excludePlHd);
}

template <const int N>
YsArray <YsPair <YsShellPolygonHandle,double> > FsGui3DMainCanvas::BoxedPlHd(int x0,int y0,int x1,int y1,YsArray <YsShellPolygonHandle,N> &excludePlHd) const
{
	return BoxedPlHd(x0,y0,x1,y1,excludePlHd.GetN(),excludePlHd);
}

template <const int N>
YsShellExt::ConstEdgeHandle FsGui3DMainCanvas::PickedCeHd(int mx,int my,YsArray <YsShellExt::ConstEdgeHandle,N> &excludeCeHd) const
{
	return PickedCeHd(mx,my,excludeCeHd.GetN(),excludeCeHd);
}

template <const int N>
YsShellExt::FaceGroupHandle FsGui3DMainCanvas::PickedFgHd(int mx,int my,YsArray <YsShellExt::FaceGroupHandle,N> &excludeFgHd) const
{
	return PickedFgHd(mx,my,excludeFgHd.GetN(),excludeFgHd);
}

template <const int N>
YsArray <YsPair <YsShellExt::ConstEdgeHandle,double> > FsGui3DMainCanvas::BoxedCeHd(int x0,int y0,int x1,int y1,const YsArray <YsShellExt::ConstEdgeHandle,N> &excludeCeHd) const
{
	return BoxedCeHd(x0,y0,x1,y1,excludeCeHd.GetN(),excludeCeHd);
}

template <const int N>
const YsShellDnmContainer <YsShellExtEdit>::Node *FsGui3DMainCanvas::PickedSlHd(int mx,int my,YsArray <const YsShellDnmContainer <YsShellExtEdit>::Node *> &excludeSlHd) const
{
	return PickedSlHd(mx,my,excludeSlHd.GetN(),excludeSlHd);
}

template <const int N>
YsShellDnmContainer <YsShellExtEdit>::Node *FsGui3DMainCanvas::PickedSlHd(int mx,int my,YsArray <const YsShellDnmContainer <YsShellExtEdit>::Node *> &excludeSlHd)
{
	return PickedSlHd(mx,my,excludeSlHd.GetN(),excludeSlHd);
}

/* } */
#endif
