#include "ysshellext_geomcalc.h"
#include <ysshellext.h>

const double YsShellExt_CalculateDihedralAngle(const YsShellExt &shl,const YsShellPolygonHandle plHd[2])
{
	return YsShellExt_CalculateDihedralAngle(shl,plHd[0],plHd[1]);
}

const double YsShellExt_CalculateDihedralAngle(const YsShellExt &shl,YsShellPolygonHandle plHd0,YsShellPolygonHandle plHd1)
{
	YsArray <YsVec3,4> plVtPos[2];
	shl.GetPolygon(plVtPos[0],plHd0);
	shl.GetPolygon(plVtPos[1],plHd1);

	YsVec3 nom[2];
	if(YSOK==YsGetAverageNormalVector(nom[0],plVtPos[0]) && YSOK==YsGetAverageNormalVector(nom[1],plVtPos[1]))
	{
		return acos(YsBound(nom[0]*nom[1],-1.0,1.0));
	}
	return 0.0;
}



const double YsShellExt_CalculateDihedralAngle(const YsShellExt &shl,const YsShellVertexHandle edVtHd[2])
{
	return YsShellExt_CalculateDihedralAngle(shl,edVtHd[0],edVtHd[1]);
}

const double YsShellExt_CalculateDihedralAngle(const class YsShellExt &shl,const YsShellVertexHandle edVtHd0,const YsShellVertexHandle edVtHd1)
{
	YsArray <YsShellPolygonHandle,2> edPlHd;
	if(YSOK==shl.FindPolygonFromEdgePiece(edPlHd,edVtHd0,edVtHd1) && 2==edPlHd.GetN())
	{
		return YsShellExt_CalculateDihedralAngle(shl,edPlHd);
	}
	return 0.0;
}

const double YsShellExt_CalculateDihedralAngle(const class YsShellExt &shl,const YsShellPolygonHandle plHd,int nThEdgeZeroBased)
{
	YsArray <YsShellVertexHandle,4> plVtHd;
	shl.GetPolygon(plVtHd,plHd);
	return YsShellExt_CalculateDihedralAngle(shl,plVtHd[nThEdgeZeroBased],plVtHd.GetCyclic(nThEdgeZeroBased+1));
}



const double YsShellExt_CalculateTotalDihedralAngleAroundEdge(const class YsShellExt &shl,const YsShellVertexHandle edVtHd[2])
{
	return YsShellExt_CalculateTotalDihedralAngleAroundEdge(shl,edVtHd[0],edVtHd[1]);
}

const double YsShellExt_CalculateTotalDihedralAngleAroundEdge(const class YsShellExt &shl,const YsShellVertexHandle edVtHd0,const YsShellVertexHandle edVtHd1)
{
	YsArray <YsShellPolygonHandle,2> basePlHdPair;
	if(YSOK==shl.FindPolygonFromEdgePiece(basePlHdPair,edVtHd0,edVtHd1) && 2==basePlHdPair.GetN())
	{
		double total=YsShellExt_CalculateDihedralAngle(shl,basePlHdPair);
		for(auto plHd : basePlHdPair)
		{
			YsArray <YsShellVertexHandle,4> plVtHd;
			shl.GetPolygon(plVtHd,plHd);
			for(auto vtIdx : plVtHd.AllIndex())
			{
				if(YSTRUE!=YsSameEdge(plVtHd[vtIdx],plVtHd.GetCyclic(vtIdx+1),edVtHd0,edVtHd1))
				{
					total+=YsShellExt_CalculateDihedralAngle(shl,plVtHd[vtIdx],plVtHd.GetCyclic(vtIdx+1));
				}
			}
		}
		return total;
	}
	return 0.0;
}

const YsVec3 YsShell_CalculatePolygonAverageNormal(const YsShell &shl,YSSIZE_T nPl,const YsShellPolygonHandle plHdArray[])
{
	YsArray <YsVec3,16> allVtPos;
	for(YSSIZE_T plIdx=0; plIdx<nPl; ++plIdx)
	{
		YsArray <YsShellVertexHandle,4> plVtHd;
		shl.GetPolygon(plVtHd,plHdArray[plIdx]);
		for(auto vtHd : plVtHd)
		{
			allVtPos.Append(shl.GetVertexPosition(vtHd));
		}
	}
	YsVec3 nom;
	if(YSOK==YsFindLeastSquareFittingPlaneNormal(nom,allVtPos))
	{
		return nom;
	}
	return YsOrigin();
}

