/* ////////////////////////////////////////////////////////////

  File Name: ysshellextedit_localop.cpp
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#include "ysshellextedit_localop.h"
#include "ysshellext_localop.h"
#include "ysshellext_trackingutil.h"

YSRESULT YsShellExtEdit_InsertVertexOnEdge(YsShellExtEdit &shl,const YsShellVertexHandle edVtHd[2],YsShellVertexHandle vtHd)
{
	YsArray <YsShellPolygonHandle,2> plHdArray;
	if(YSOK==shl.FindPolygonFromEdgePiece(plHdArray,edVtHd))
	{
		for(YSSIZE_T plIdx=0; plIdx<plHdArray.GetN(); ++plIdx)
		{
			YsArray <YsShellVertexHandle,4> plVtHd;
			shl.GetVertexListOfPolygon(plVtHd,plHdArray[plIdx]);
			if(YSOK==YsShellExt_InsertVertexOnEdgeOfVertexArray(plVtHd,YSTRUE,edVtHd,vtHd))
			{
				shl.SetPolygonVertex(plHdArray[plIdx],plVtHd);
			}
		}
	}

	YsArray <YsShellExt::ConstEdgeHandle,2> ceHdArray;
	if(YSOK==shl.FindConstEdgeFromEdgePiece(ceHdArray,edVtHd))
	{
		for(YSSIZE_T ceIdx=0; ceIdx<ceHdArray.GetN(); ++ceIdx)
		{
			YSBOOL isLoop;
			YsArray <YsShellVertexHandle,16> ceVtHd;
			shl.GetConstEdge(ceVtHd,isLoop,ceHdArray[ceIdx]);
			if(YSOK==YsShellExt_InsertVertexOnEdgeOfVertexArray(ceVtHd,isLoop,edVtHd,vtHd))
			{
				shl.ModifyConstEdge(ceHdArray[ceIdx],ceVtHd,isLoop);
			}
		}
	}

	return YSOK;
}

YSRESULT YsShellExtEdit_CollapseEdge(YsShellExtEdit &shl,YsShellVertexHandle vtHdToDelete,YsShellVertexHandle vtHdToSurvive)
{
	YsShell_CollapseInfo info;
	if(YSOK==info.MakeInfo((const YsShellExt &)shl,vtHdToDelete,vtHdToSurvive))
	{
		for(YSSIZE_T idx=0; idx<info.newPlg.GetN(); ++idx)
		{
			shl.SetPolygonVertex(info.newPlg[idx].plHd,info.newPlg[idx].plVtHd);
		}
		for(YSSIZE_T idx=0; idx<info.newConstEdge.GetN(); ++idx)
		{
			shl.ModifyConstEdge(info.newConstEdge[idx].ceHd,info.newConstEdge[idx].ceVtHd,info.newConstEdge[idx].isLoop);
		}
		shl.ForceDeleteMultiPolygon(info.plHdToDel);
		shl.DeleteMultiConstEdge(info.ceHdToDel);
		shl.DeleteVertex(vtHdToDelete);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExtEdit_TriangulatePolygon(YsShellExtEdit &shl,YsShellPolygonHandle plHd)
{
	YsArray <YsShellVertexHandle,4> plVtHd;
	shl.GetPolygon(plVtHd,plHd);

	YsSword sword;
	YsArray <YsVec3,4> plVtPos;
	YsArray <int,4> plVtIdx;

	plVtPos.Set(plVtHd.GetN(),NULL);
	plVtIdx.Set(plVtHd.GetN(),NULL);
	for(auto idx : plVtHd.AllIndex())
	{
		shl.GetVertexPosition(plVtPos[idx],plVtHd[idx]);
		plVtIdx[idx]=idx;
	}

	sword.SetInitialPolygon(plVtPos.GetN(),plVtPos,plVtIdx);
	if(YSOK==sword.Triangulate())
	{
		YsShellExtEdit::StopIncUndo incUndo(shl);
		for(int triIdx=0; triIdx<sword.GetNumPolygon(); ++triIdx)
		{
			const YsArray <int> *tri=sword.GetVertexIdList(triIdx);
			if(NULL!=tri && 3==tri->GetN())
			{
				const YsShellVertexHandle triVtHd[3]=
				{
					plVtHd[plVtIdx[(*tri)[0]]],
					plVtHd[plVtIdx[(*tri)[1]]],
					plVtHd[plVtIdx[(*tri)[2]]]
				};
				if(0==triIdx)
				{
					shl.SetPolygonVertex(plHd,3,triVtHd);
				}
				else
				{
					auto newPlHd=shl.AddPolygon(3,triVtHd);

					auto attrib=shl.GetPolygonAttrib(plHd);
					shl.SetPolygonAttrib(newPlHd,*attrib);

					auto nom=shl.GetNormal(plHd);
					auto col=shl.GetColor(plHd);

					shl.SetPolygonNormal(newPlHd,nom);
					shl.SetPolygonColor(newPlHd,col);
				}
			}
		}

		return YSOK;
	}
	return YSERR;
}

YsShellExt::ConstEdgeHandle YsShellExtEdit_SplitConstEdge(YsShellExtEdit &shl,YsShellExt::ConstEdgeHandle ceHd,YsShellVertexHandle vtHd)
{
	YsShellExtEdit::StopIncUndo incUndo(shl);

	YSBOOL isLoop;
	YsArray <YsShellVertexHandle,16> ceVtHd;
	shl.GetConstEdge(ceVtHd,isLoop,ceHd);
	if(YSTRUE!=isLoop)
	{
		for(YSSIZE_T idx=1; idx<ceVtHd.GetN()-1; ++idx)
		{
			if(ceVtHd[idx]==vtHd)
			{
				YsArray <YsShellVertexHandle,16> newCeVtHd;
				newCeVtHd.Set(ceVtHd.GetN()-idx,ceVtHd.GetArray()+idx);
				auto attrib=shl.GetConstEdgeAttrib(ceHd);

				auto newCeHd=shl.AddConstEdge(newCeVtHd,YSFALSE);
				shl.SetConstEdgeAttrib(newCeHd,attrib);

				shl.ModifyConstEdge(ceHd,idx+1,ceVtHd);

				return newCeHd;
			}
		}
	}
	else
	{
		for(auto idx : ceVtHd.AllIndex())
		{
			if(ceVtHd[idx]==vtHd)
			{
				for(int copyIdx=0; copyIdx<=idx; ++copyIdx)
				{
					ceVtHd.Append(ceVtHd[copyIdx]);
				}
				ceVtHd.Delete(0,idx);
				shl.ModifyConstEdge(ceHd,ceVtHd,YSFALSE);
				return ceHd;
			}
		}
	}
	return NULL;
}

YsShellExt::ConstEdgeHandle YsShellExtEdit_MergeConstEdge(YsShellExtEdit &shl,YSSIZE_T nCe,const YsShellExt::ConstEdgeHandle ceHdArray[],YSBOOL makeLoopIfClosed)
{
	auto vtHdChain=YsShellExt_TrackingUtil::MakeVertexSequenceFromMultiConstEdge((const YsShellExt &)shl,nCe,ceHdArray);
	if(2>vtHdChain.GetN())
	{
		return NULL;
	}

	YsShellExtEdit::StopIncUndo incUndo(shl);
	YSBOOL isLoop=YSFALSE;
	if(YSTRUE==makeLoopIfClosed && 4<=vtHdChain.GetN() && vtHdChain[0]==vtHdChain.Last())
	{
		isLoop=YSTRUE;
		vtHdChain.DeleteLast();
	}

	shl.ModifyConstEdge(ceHdArray[0],vtHdChain,isLoop);
	for(YSSIZE_T idx=1; idx<nCe; ++idx)
	{
		if(ceHdArray[idx]!=ceHdArray[0])
		{
			shl.DeleteConstEdge(ceHdArray[idx]);
		}
	}

	return ceHdArray[0];
}

YSRESULT YsShellExtEdit_ApplyPolygonSplit(YsShellExtEdit &shl,const YsShell_LocalOperation::PolygonSplit &plgSplit)
{
	YSBOOL first=YSTRUE;
	for(auto &newPlVtHd : plgSplit.plVtHdArray)
	{
		if(YSTRUE==first)
		{
			shl.SetPolygonVertex(plgSplit.plHd,newPlVtHd);
			first=YSFALSE;
		}
		else
		{
			auto nom=shl.GetNormal(plgSplit.plHd);
			auto col=shl.GetColor(plgSplit.plHd);
			auto attrib=shl.GetPolygonAttrib(plgSplit.plHd);
			auto fgHd=shl.FindFaceGroupFromPolygon(plgSplit.plHd);

			auto newPlHd=shl.AddPolygon(newPlVtHd);
			shl.SetPolygonNormal(newPlHd,nom);
			shl.SetPolygonColor(newPlHd,col);
			shl.SetPolygonAttrib(newPlHd,*attrib);
			if(NULL!=fgHd)
			{
				shl.AddFaceGroupPolygon(fgHd,1,&newPlHd);
			}
		}
	}
	return YSOK;
}

YSRESULT YsShellExtEdit_ApplyConstEdgeSplit(YsShellExtEdit &shl,const YsShell_LocalOperation::ConstEdgeSplit &ceSplit)
{
	YSBOOL first=YSTRUE;
	for(auto &newCeVtHd : ceSplit.ceVtHdArray)
	{
		if(YSTRUE==first)
		{
			shl.ModifyConstEdge(ceSplit.ceHd,newCeVtHd,YSFALSE);
			first=YSFALSE;
		}
		else
		{
			auto attrib=shl.GetConstEdgeAttrib(ceSplit.ceHd);
			auto newCeHd=shl.AddConstEdge(newCeVtHd,YSFALSE);
			shl.SetConstEdgeAttrib(newCeHd,attrib);
		}
	}
	return YSOK;
}

YSRESULT YsShellExtEdit_ApplyFaceGroupSplit(YsShellExtEdit &shl,const YsShell_LocalOperation::FaceGroupSplit &fgSplit)
{
	if(2<=fgSplit.fgPlHdArray.GetN())
	{
		shl.ModifyFaceGroup(fgSplit.fgHd,fgSplit.fgPlHdArray[0]);
		for(auto index : fgSplit.fgPlHdArray.AllIndex())
		{
			if(0<index)
			{
				auto &fgPlHd=fgSplit.fgPlHdArray[index];
				auto newFgHd=shl.AddFaceGroup(fgPlHd);
				auto attrib=shl.GetFaceGroupAttrib(fgSplit.fgHd);
				shl.SetFaceGroupAttrib(newFgHd,attrib);
			}
		}
		return YSOK;
	}
	return YSERR;
}

