/* ////////////////////////////////////////////////////////////

  File Name: ysshellextedit_localop.h
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#ifndef YSSHELLEXTEDIT_LOCALOP_IS_INCLUDED
#define YSSHELLEXTEDIT_LOCALOP_IS_INCLUDED
/* { */

#include <ysshellextedit.h>
#include "ysshellext_localop.h"

YSRESULT YsShellExtEdit_InsertVertexOnEdge(YsShellExtEdit &shl,const YsShellVertexHandle edVtHd[2],YsShellVertexHandle vtHd);
YSRESULT YsShellExtEdit_CollapseEdge(YsShellExtEdit &shl,YsShellVertexHandle vtHdToDelete,YsShellVertexHandle vtHdToSurvive);

YSRESULT YsShellExtEdit_TriangulatePolygon(YsShellExtEdit &shl,YsShellPolygonHandle plHd);

/*! This function splits a constraint edge ceHd at vertex vtHd, and returns a handle to the new const edge.
    If constraint edge ceHd does not include vertex vtHd, this function will fail (do nothing) and returns NULL.
    If constraint edge ceHd is an open constraint edge, and vtHd is the first or the last of its vertices, this functil will fail and returns NULL.
    If constraint edge ceHd is a loop, it will be made an open constraint edge that has the same first and the last vertex, which is vtHd, and the
    return value will also be ceHd. */
YsShellExt::ConstEdgeHandle YsShellExtEdit_SplitConstEdge(YsShellExtEdit &shl,YsShellExt::ConstEdgeHandle ceHd,YsShellVertexHandle vtHd);

/*! This function merges given constraint edges into one continuous constraint edge.
    If successful, this function will return the handle to the newly assembled const-edge, which will be one of the given handles.
    If makeLoopIfClosed is YSTRUE, and if chained const-edges form a loop, const-edges will become a loop.
    If it cannot make a continuous vertex chain without a branch from the given const edges, this function will fail (do nothing) and return NULL. */
YsShellExt::ConstEdgeHandle YsShellExtEdit_MergeConstEdge(YsShellExtEdit &shl,YSSIZE_T nCe,const YsShellExt::ConstEdgeHandle ceHdArray[],YSBOOL makeLoopIfClosed);

template <const int N>
YsShellExt::ConstEdgeHandle YsShellExtEdit_MergeConstEdge(YsShellExtEdit &shl,const YsArray <YsShellExt::ConstEdgeHandle,N> &ceHdArray,YSBOOL makeLoopIfClosed)
{
	return YsShellExtEdit_MergeConstEdge(shl,ceHdArray.GetN(),ceHdArray,makeLoopIfClosed);
}


/*! This function applies the changes desribed in YsShell_LocalOperation::PolygonSplit. */
YSRESULT YsShellExtEdit_ApplyPolygonSplit(YsShellExtEdit &shl,const YsShell_LocalOperation::PolygonSplit &plgSplit);

/*! This function applies the changes desribed in YsShell_LocalOperation::ConstEdgeSplit. */
YSRESULT YsShellExtEdit_ApplyConstEdgeSplit(YsShellExtEdit &shl,const YsShell_LocalOperation::ConstEdgeSplit &ceSplit);

/*! This function applies the changes described in YsShell_LocalOperation::FaceGroupSplit. */
YSRESULT YsShellExtEdit_ApplyFaceGroupSplit(YsShellExtEdit &shl,const YsShell_LocalOperation::FaceGroupSplit &ceSplit);

/* } */
#endif
