#ifndef YSGLBITMAP_IS_INCLUDED
#define YSGLBITMAP_IS_INCLUDED
/* { */

#include "ysgldef.h"

#ifdef __cplusplus
extern "C" {
#endif

/*! This function creates a copy of a bitmap.
    The destination bitmap must be the same size.
 */
void YsGLCopyBitmap(int w,int h,GLfloat *dst,const GLfloat *src);

/*! This function converts a unsigned char RGBA bitmap to GLfloat bitmap. */
void YsGLCopyUCharBitmapToGLFloatBitmap(int w,int h,GLfloat *dst,const unsigned char *src);

/*! This function virtually maps the bitmap on a sphere and calculates the lighting.
    All pixels outside the spheres will be given color of (0,0,0,0)

    This function only supports directional light and does not support point light.
    Fourth component of light position will be ignored.

    When useSpecular is non zero, specular reflection will be calculated by assuming that the sphere is infinitely far in a view coordinate.

    ambientColor must be of length of 3.
    lightEnable must be of length of nLight.
    lightPosition must be of length of 4*nLight.  (Although each 4th component will be ignored.)
    lightColor must be of length of 3*nLight.
    specularColor must be of length of 3.
*/
void YsGLApplySphereMap(
    int w,int h,GLfloat *rgba,
    GLfloat ambientColor[],  // r,g,b
    int nLight,
	GLint lightEnabled[],
	GLfloat lightPosition[], // x,y,z,w
	GLfloat lightColor[],    // r,g,b
	int useSpecular,
	GLfloat specularColor[], // r,g,b
	GLfloat specularExponent);

#ifdef __cplusplus
}
#endif

/* } */
#endif
